"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.indexGtf = void 0;
const tslib_1 = require("tslib");
const util_1 = require("../util");
const cli_progress_1 = require("cli-progress");
const zlib_1 = require("zlib");
const readline_1 = (0, tslib_1.__importDefault)(require("readline"));
function indexGtf(config, attributes, outLocation, typesToExclude, quiet) {
    return (0, tslib_1.__asyncGenerator)(this, arguments, function* indexGtf_1() {
        var e_1, _a;
        const { adapter, trackId } = config;
        const { gtfLocation: { uri }, } = adapter;
        // progress bar code was aided by blog post at
        // https://webomnizz.com/download-a-file-with-progressbar-using-node-js/
        const progressBar = new cli_progress_1.SingleBar({
            format: '{bar} ' + trackId + ' {percentage}% | ETA: {eta}s',
            etaBuffer: 2000,
        }, cli_progress_1.Presets.shades_classic);
        let receivedBytes = 0;
        const { totalBytes, stream } = yield (0, tslib_1.__await)((0, util_1.getLocalOrRemoteStream)(uri, outLocation));
        if (!quiet) {
            progressBar.start(totalBytes, 0);
        }
        stream.on('data', chunk => {
            receivedBytes += chunk.length;
            progressBar.update(receivedBytes);
        });
        const gzStream = uri.match(/.b?gz$/) ? stream.pipe((0, zlib_1.createGunzip)()) : stream;
        const rl = readline_1.default.createInterface({
            input: gzStream,
        });
        try {
            for (var rl_1 = (0, tslib_1.__asyncValues)(rl), rl_1_1; rl_1_1 = yield (0, tslib_1.__await)(rl_1.next()), !rl_1_1.done;) {
                const line = rl_1_1.value;
                if (line.startsWith('#')) {
                    continue;
                }
                else if (line.startsWith('>')) {
                    break;
                }
                const [seq_name, , type, start, end, , , , col9] = line.split('\t');
                const locStr = `${seq_name}:${start}..${end}`;
                if (!typesToExclude.includes(type)) {
                    const col9Attrs = Object.fromEntries(col9
                        .split(';')
                        .map(f => f.trim())
                        .filter(f => !!f)
                        .map(f => f.split(' '))
                        .map(([key, val]) => {
                        return [
                            key.trim(),
                            val.trim().split(',').join(' ').replace(/("|')/g, ''),
                        ];
                    }));
                    const attrs = attributes
                        .map(attr => col9Attrs[attr])
                        .filter((f) => !!f);
                    if (attrs.length) {
                        const record = JSON.stringify([
                            encodeURIComponent(locStr),
                            encodeURIComponent(trackId),
                            ...attrs.map(a => encodeURIComponent(a || '')),
                        ]).replace(/,/g, '|');
                        yield yield (0, tslib_1.__await)(`${record} ${[...new Set(attrs)].join(' ')}\n`);
                    }
                }
            }
        }
        catch (e_1_1) { e_1 = { error: e_1_1 }; }
        finally {
            try {
                if (rl_1_1 && !rl_1_1.done && (_a = rl_1.return)) yield (0, tslib_1.__await)(_a.call(rl_1));
            }
            finally { if (e_1) throw e_1.error; }
        }
        progressBar.stop();
    });
}
exports.indexGtf = indexGtf;
